#!/usr/bin/env python3
#  Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

"""Access the results of benchmark runs
Send these results on to OpenSearch graphing service
"""

import argparse
import itertools
import logging
import os
import re
import sys

import requests
from dateutil import parser

logging.basicConfig(level=logging.DEBUG)


class Configuration:
    opensearch_user = os.environ["ES_USER"]
    opensearch_pass = os.environ["ES_PASS"]


class BenchmarkResultException(Exception):
    def __init__(self, message, content):
        super().__init__(self, message)
        self.content = content


class BenchmarkUtils:

    expected_keys = [
        "ops_sec",
        "mb_sec",
        "lsm_sz",
        "blob_sz",
        "c_wgb",
        "w_amp",
        "c_mbps",
        "c_wsecs",
        "c_csecs",
        "b_rgb",
        "b_wgb",
        "usec_op",
        "p50",
        "p99",
        "p99.9",
        "p99.99",
        "pmax",
        "uptime",
        "stall%",
        "Nstall",
        "u_cpu",
        "s_cpu",
        "rss",
        "test",
        "date",
        "version",
        "job_id",
    ]

    def sanity_check(row):
        if "test" not in row:
            logging.debug(f"not 'test' in row: {row}")
            return False
        if row["test"] == "":
            logging.debug(f"row['test'] == '': {row}")
            return False
        if "date" not in row:
            logging.debug(f"not 'date' in row: {row}")
            return False
        if "ops_sec" not in row:
            logging.debug(f"not 'ops_sec' in row: {row}")
            return False
        try:
            _ = int(row["ops_sec"])
        except (ValueError, TypeError):
            logging.debug(f"int(row['ops_sec']): {row}")
            return False
        try:
            (_, _) = parser.parse(row["date"], fuzzy_with_tokens=True)
        except (parser.ParserError):
            logging.error(
                f"parser.parse((row['date']): not a valid format for date in row: {row}"
            )
            return False
        return True

    def conform_opensearch(row):
        (dt, _) = parser.parse(row["date"], fuzzy_with_tokens=True)
        # create a test_date field, which was previously what was expected
        # repair the date field, which has what can be a WRONG ISO FORMAT, (no leading 0 on single-digit day-of-month)
        # e.g. 2022-07-1T00:14:55 should be 2022-07-01T00:14:55
        row["test_date"] = dt.isoformat()
        row["date"] = dt.isoformat()
        return {key.replace(".", "_"): value for key, value in row.items()}


class ResultParser:
    def __init__(self, field=r"(\w|[+-:.%])+", intrafield=r"(\s)+", separator="\t"):
        self.field = re.compile(field)
        self.intra = re.compile(intrafield)
        self.sep = re.compile(separator)

    def ignore(self, l_in: str):
        if len(l_in) == 0:
            return True
        if l_in[0:1] == "#":
            return True
        return False

    def line(self, line_in: str):
        """Parse a line into items
        Being clever about separators
        """
        line = line_in
        row = []
        while line != "":
            match_item = self.field.match(line)
            if match_item:
                item = match_item.group(0)
                row.append(item)
                line = line[len(item) :]
            else:
                match_intra = self.intra.match(line)
                if match_intra:
                    intra = match_intra.group(0)
                    # Count the separators
                    # If there are >1 then generate extra blank fields
                    # White space with no true separators fakes up a single separator
                    tabbed = self.sep.split(intra)
                    sep_count = len(tabbed) - 1
                    if sep_count == 0:
                        sep_count = 1
                    for _ in range(sep_count - 1):
                        row.append("")
                    line = line[len(intra) :]
                else:
                    raise BenchmarkResultException(
                        "Invalid TSV line", f"{line_in} at {line}"
                    )
        return row

    def parse(self, lines):
        """Parse something that iterates lines"""
        rows = [self.line(line) for line in lines if not self.ignore(line)]
        header = rows[0]
        width = len(header)
        records = [
            {k: v for (k, v) in itertools.zip_longest(header, row[:width])}
            for row in rows[1:]
        ]
        return records


def load_report_from_tsv(filename: str):
    file = open(filename)
    contents = file.readlines()
    file.close()
    parser = ResultParser()
    report = parser.parse(contents)
    logging.debug(f"Loaded TSV Report: {report}")
    return report


def push_report_to_opensearch(report, esdocument):
    sanitized = [
        BenchmarkUtils.conform_opensearch(row)
        for row in report
        if BenchmarkUtils.sanity_check(row)
    ]
    logging.debug(
        f"upload {len(sanitized)} sane of {len(report)} benchmarks to opensearch"
    )
    for single_benchmark in sanitized:
        logging.debug(f"upload benchmark: {single_benchmark}")
        response = requests.post(
            esdocument,
            json=single_benchmark,
            auth=(os.environ["ES_USER"], os.environ["ES_PASS"]),
        )
        logging.debug(
            f"Sent to OpenSearch, status: {response.status_code}, result: {response.text}"
        )
        response.raise_for_status()


def push_report_to_null(report):

    for row in report:
        if BenchmarkUtils.sanity_check(row):
            logging.debug(f"row {row}")
            conformed = BenchmarkUtils.conform_opensearch(row)
            logging.debug(f"conformed row {conformed}")


def main():
    """Tool for fetching, parsing and uploading benchmark results to OpenSearch / ElasticSearch
    This tool will

    (1) Open a local tsv benchmark report file
    (2) Upload to OpenSearch document, via https/JSON
    """

    parser = argparse.ArgumentParser(description="CircleCI benchmark scraper.")

    # --tsvfile is the name of the file to read results from
    # --esdocument is the ElasticSearch document to push these results into
    #
    parser.add_argument(
        "--tsvfile",
        default="build_tools/circle_api_scraper_input.txt",
        help="File from which to read tsv report",
    )
    parser.add_argument(
        "--esdocument",
        help="ElasticSearch/OpenSearch document URL to upload report into",
    )
    parser.add_argument(
        "--upload", choices=["opensearch", "none"], default="opensearch"
    )

    args = parser.parse_args()
    logging.debug(f"Arguments: {args}")
    reports = load_report_from_tsv(args.tsvfile)
    if args.upload == "opensearch":
        push_report_to_opensearch(reports, args.esdocument)
    else:
        push_report_to_null(reports)


if __name__ == "__main__":
    sys.exit(main())



#  Copyright (c) 2011-present, Facebook, Inc.  All rights reserved.
#  This source code is licensed under both the GPLv2 (found in the
#  COPYING file in the root directory) and Apache 2.0 License
#  (found in the LICENSE.Apache file in the root directory).

"""Filter for error messages in test output:
    - Receives merged stdout/stderr from test on stdin
    - Finds patterns of known error messages for test name (first argument)
    - Prints those error messages to stdout
"""


import re
import sys


class ErrorParserBase:
    def parse_error(self, line):
        """Parses a line of test output. If it contains an error, returns a
        formatted message describing the error; otherwise, returns None.
        Subclasses must override this method.
        """
        raise NotImplementedError


class GTestErrorParser(ErrorParserBase):
    """A parser that remembers the last test that began running so it can print
    that test's name upon detecting failure.
    """

    _GTEST_NAME_PATTERN = re.compile(r"\[ RUN      \] (\S+)$")
    # format: '<filename or "unknown file">:<line #>: Failure'
    _GTEST_FAIL_PATTERN = re.compile(r"(unknown file|\S+:\d+): Failure$")

    def __init__(self):
        self._last_gtest_name = "Unknown test"

    def parse_error(self, line):
        gtest_name_match = self._GTEST_NAME_PATTERN.match(line)
        if gtest_name_match:
            self._last_gtest_name = gtest_name_match.group(1)
            return None
        gtest_fail_match = self._GTEST_FAIL_PATTERN.match(line)
        if gtest_fail_match:
            return "{} failed: {}".format(self._last_gtest_name, gtest_fail_match.group(1))
        return None


class MatchErrorParser(ErrorParserBase):
    """A simple parser that returns the whole line if it matches the pattern."""

    def __init__(self, pattern):
        self._pattern = re.compile(pattern)

    def parse_error(self, line):
        if self._pattern.match(line):
            return line
        return None


class CompilerErrorParser(MatchErrorParser):
    def __init__(self):
        # format (compile error):
        #   '<filename>:<line #>:<column #>: error: <error msg>'
        # format (link error):
        #   '<filename>:<line #>: error: <error msg>'
        # The below regex catches both
        super().__init__(r"\S+:\d+: error:")


class ScanBuildErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(r"scan-build: \d+ bugs found.$")


class DbCrashErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(r"\*\*\*.*\^$|TEST FAILED.")


class WriteStressErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(
            r"ERROR: write_stress died with exitcode=\d+"
        )


class AsanErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(r"==\d+==ERROR: AddressSanitizer:")


class UbsanErrorParser(MatchErrorParser):
    def __init__(self):
        # format: '<filename>:<line #>:<column #>: runtime error: <error msg>'
        super().__init__(r"\S+:\d+:\d+: runtime error:")


class ValgrindErrorParser(MatchErrorParser):
    def __init__(self):
        # just grab the summary, valgrind doesn't clearly distinguish errors
        # from other log messages.
        super().__init__(r"==\d+== ERROR SUMMARY:")


class CompatErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(r"==== .*[Ee]rror.* ====$")


class TsanErrorParser(MatchErrorParser):
    def __init__(self):
        super().__init__(r"WARNING: ThreadSanitizer:")


_TEST_NAME_TO_PARSERS = {
    "punit": [CompilerErrorParser, GTestErrorParser],
    "unit": [CompilerErrorParser, GTestErrorParser],
    "release": [CompilerErrorParser, GTestErrorParser],
    "unit_481": [CompilerErrorParser, GTestErrorParser],
    "release_481": [CompilerErrorParser, GTestErrorParser],
    "clang_unit": [CompilerErrorParser, GTestErrorParser],
    "clang_release": [CompilerErrorParser, GTestErrorParser],
    "clang_analyze": [CompilerErrorParser, ScanBuildErrorParser],
    "code_cov": [CompilerErrorParser, GTestErrorParser],
    "unity": [CompilerErrorParser, GTestErrorParser],
    "lite": [CompilerErrorParser],
    "lite_test": [CompilerErrorParser, GTestErrorParser],
    "stress_crash": [CompilerErrorParser, DbCrashErrorParser],
    "stress_crash_with_atomic_flush": [CompilerErrorParser, DbCrashErrorParser],
    "stress_crash_with_txn": [CompilerErrorParser, DbCrashErrorParser],
    "write_stress": [CompilerErrorParser, WriteStressErrorParser],
    "asan": [CompilerErrorParser, GTestErrorParser, AsanErrorParser],
    "asan_crash": [CompilerErrorParser, AsanErrorParser, DbCrashErrorParser],
    "asan_crash_with_atomic_flush": [
        CompilerErrorParser,
        AsanErrorParser,
        DbCrashErrorParser,
    ],
    "asan_crash_with_txn": [CompilerErrorParser, AsanErrorParser, DbCrashErrorParser],
    "ubsan": [CompilerErrorParser, GTestErrorParser, UbsanErrorParser],
    "ubsan_crash": [CompilerErrorParser, UbsanErrorParser, DbCrashErrorParser],
    "ubsan_crash_with_atomic_flush": [
        CompilerErrorParser,
        UbsanErrorParser,
        DbCrashErrorParser,
    ],
    "ubsan_crash_with_txn": [CompilerErrorParser, UbsanErrorParser, DbCrashErrorParser],
    "valgrind": [CompilerErrorParser, GTestErrorParser, ValgrindErrorParser],
    "tsan": [CompilerErrorParser, GTestErrorParser, TsanErrorParser],
    "format_compatible": [CompilerErrorParser, CompatErrorParser],
    "run_format_compatible": [CompilerErrorParser, CompatErrorParser],
    "no_compression": [CompilerErrorParser, GTestErrorParser],
    "run_no_compression": [CompilerErrorParser, GTestErrorParser],
    "regression": [CompilerErrorParser],
    "run_regression": [CompilerErrorParser],
}


def main():
    if len(sys.argv) != 2:
        return "Usage: %s <test name>" % sys.argv[0]
    test_name = sys.argv[1]
    if test_name not in _TEST_NAME_TO_PARSERS:
        return "Unknown test name: %s" % test_name

    error_parsers = []
    for parser_cls in _TEST_NAME_TO_PARSERS[test_name]:
        error_parsers.append(parser_cls())

    for line in sys.stdin:
        line = line.strip()
        for error_parser in error_parsers:
            error_msg = error_parser.parse_error(line)
            if error_msg is not None:
                print(error_msg)


if __name__ == "__main__":
    sys.exit(main())
